"""NumSpace provides the namespace class for the enumerations."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..mcls import AbstractNamespace
from ..parse import maybe
from . import NumHook

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from worktoy.keenum import NUM


class NumSpace(AbstractNamespace):
  """NumSpace provides the namespace class for the enumerations."""

  __member_nums__ = None

  def getMemberNums(self, ) -> list[NUM]:
    """Get the member nums of the enumeration."""
    return maybe(self.__member_nums__, [])

  def addNum(self, key: str, num: NUM) -> bool:
    """Add a num to the enumeration."""
    existing = self.getMemberNums()
    num.key = key
    self.__member_nums__ = [*existing, num]
    return True

  numHook = NumHook()



"""UnrecognizedMember is raised when a KeeNum class is unabled to
recognize a given identifier."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..text import monoSpace

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from ..keenum import MetaNum
  from typing import Any


class UnrecognizedMember(Exception):
  """UnrecognizedMember is raised when a KeeNum class is unable to
  recognize a given identifier."""

  keenumCls = _Attribute()
  unrecognizedIdentifier = _Attribute()

  def __init__(self, cls: MetaNum, identifier: Any) -> None:
    """Initialize the UnrecognizedMember object."""
    self.keenumCls = cls
    self.unrecognizedIdentifier = identifier
    infoSpec = """KeeNum class '%s' could not recognize identifier '%s'!"""
    info = infoSpec % (cls.__name__, identifier)
    Exception.__init__(self, monoSpace(info))

  def _resolveOther(self, other: Any) -> UnrecognizedMember:
    """Resolve the other object."""
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, (tuple, list)):
      try:
        return cls(*other)
      except TypeError:
        return NotImplemented
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Compare the UnrecognizedMember object with another object."""
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    cls = type(self)
    if isinstance(other, cls):
      if self.keenumCls != other.keenumCls:
        return False
      if self.unrecognizedIdentifier != other.unrecognizedIdentifier:
        return False
      return True
    return False



"""The 'worktoy.waitaminute' module provides custom exception classes. """
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from ._attribute import _Attribute
from ._abstract_exception import AbstractException
from ._cast_mismatch import CastMismatch
from ._dispatch_exception import DispatchException
from ._resolve_exception import ResolveException
from ._missing_variable import MissingVariable
from ._variable_not_none import VariableNotNone
from ._questionable_syntax import QuestionableSyntax
from ._read_only_error import ReadOnlyError
from ._reserved_name import ReservedName
from ._hook_exception import HookException
from ._hash_mismatch import HashMismatch
from ._unrecognized_member import UnrecognizedMember
from ._write_once_error import WriteOnceError
from ._subclass_exception import SubclassException
from ._protected_error import ProtectedError
from ._type_exception import TypeException
from ._deleted_attribute_exception import DeletedAttributeException

__all__ = [
    'CastMismatch',
    'DispatchException',
    'ResolveException',
    'MissingVariable',
    'VariableNotNone',
    'QuestionableSyntax',
    'ReadOnlyError',
    'ReservedName',
    'HookException',
    'HashMismatch',
    'UnrecognizedMember',
    'WriteOnceError',
    'SubclassException',
    'ProtectedError',
    'TypeException',
]



"""CastMismatch should be raised to indicate that the fast static
system of the TypeSig class did not match."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..text import monoSpace
from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, Any


class CastMismatch(TypeError):
  """CastMismatch should be raised to indicate that the fast static
  system of the TypeSig class did not match."""

  expType = _Attribute()
  actObj = _Attribute()
  actType = _Attribute()

  def __init__(self, type_: type, obj: object) -> None:
    """Initialize the CastMismatch object."""
    self.expType = type_
    self.actObj = obj
    self.actType = type(obj)
    typeName = type_.__name__
    objStr = str(obj)
    objType = type(obj).__name__
    infoSpec = """Unable to cast object: '%s' of type '%s' to type '%s'!"""
    info = monoSpace(infoSpec % (objStr, objType, typeName))
    TypeError.__init__(self, info)

  def _resolveOther(self, other: object) -> Self:
    """Resolve the other object."""
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, (tuple, list)):
      try:
        return cls(*other)
      except TypeError:
        return NotImplemented
    if isinstance(other, dict):
      type_ = other.get('__expected_type__', None)
      obj = other.get('__actual_object__', None)
      if type_ is not None and obj is not None:
        if isinstance(type_, type):
          return cls(type_, obj)
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Check if the exception is equal to another object."""
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    if self.expType != other.expType:
      return False
    if self.actObj != other.actObj:
      return False
    if self.actType != other.actType:
      return False
    return True



"""EZSpace provides the namespace for the EZData class. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import DataField, EZHook
from ..mcls import BaseSpace
from ..parse import maybe

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False


class EZSpace(BaseSpace):
  """EZSpace provides the namespace for the EZData class. """

  __data_fields__ = None

  def getDataFields(self) -> list[DataField]:
    """Get the data fields."""
    return maybe(self.__data_fields__, [])

  def addField(self, key: str, type_: type, val: object) -> None:
    """Add a field to the data fields."""
    if self.__data_fields__ is None:
      self.__data_fields__ = []
    dataField = DataField(key, type_, val)
    existing = self.getDataFields()
    self.__data_fields__ = [*existing, dataField]

  ezHook = EZHook()



"""AbstractMetaclass provides the baseclass for custom metaclasses. """
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from ..text import monoSpace, typeMsg
from ..waitaminute import QuestionableSyntax
from . import Base
from . import AbstractNamespace as ASpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, Any, Callable


class _MetaMetaclass(type):
  """MetaMetaclass is necessary to customize the __str__ method of a
  metaclass"""

  def __str__(cls, ) -> str:
    """Returns the name of the class. """
    return """%s[metaclass=%s]""" % (cls.__name__, cls.__class__.__name__)


class AbstractMetaclass(_MetaMetaclass, metaclass=_MetaMetaclass):
  """The AbstractMetaclass class provides a base class for custom
  metaclasses."""

  @staticmethod
  def _validateNamespace(namespace: dict, **kwargs) -> dict:
    """The _validateNamespace method is invoked to validate the namespace
    object before the class is created. """
    if '__del__' in namespace and '__delete__' not in namespace:
      if not kwargs.get('trustMeBro', False):
        e = """The namespace encountered the '__del__' method! 
          This method has very limited practical use. It has significant 
          potential for unexpected behaviour. If '__del__' were actually
          the intention, the 'worktoy' library requires passing the 
          keyword 'trustMeBro=True' to the class creation."""
        raise SyntaxError(monoSpace(e))
    derpNames = [
        '__get_item__', '__set_item__', '__get_attr__',
        '__set_attr__'
    ]
    realNames = ['__getitem__', '__setitem__', '__getattr__', '__setattr__']
    for derp, name in zip(derpNames, realNames):
      if derp in namespace:
        raise QuestionableSyntax(derp, name, )
    return namespace

  @staticmethod
  def _notifySubclassHook(cls, *bases) -> type:
    """The _notifySubclassHook method is invoked to notify each baseclass
    of the created class of the class creation."""
    for base in bases:
      hook = getattr(base, '__subclasshook__', None)
      if hook is None:
        continue
      hook(cls)
    return cls

  @classmethod
  def __prepare__(mcls, name: str, bases: Base, **kwargs) -> ASpace:
    """The __prepare__ method is invoked before the class is created. This
    implementation ensures that the created class has access to the safe
    __init__ and __init_subclass__ through the BaseObject class in its
    method resolution order."""
    return ASpace(mcls, name, bases, **kwargs)

  def __new__(mcls, name: str, bases: Base, space: ASpace, **kwargs) -> type:
    """The __new__ method is invoked to create the class."""
    namespace = mcls._validateNamespace(space.compile(), **kwargs)
    cls = _MetaMetaclass.__new__(mcls, name, bases, namespace, **kwargs)
    return mcls._notifySubclassHook(cls, *bases)

  def __init__(cls, name: str, bases: Base, space: ASpace, **kwargs) -> None:
    """The __init__ method is invoked to initialize the class."""
    if TYPE_CHECKING:
      assert isinstance(space, ASpace)
      assert isinstance(bases, tuple)
    _MetaMetaclass.__init__(cls, name, bases, space, **kwargs)

  def __call__(cls, *args, **kwargs) -> Any:
    """The __call__ method is invoked when the class is called."""
    classCall = getattr(cls, '__class_call__', None)
    if classCall is None:
      return type.__call__(cls, *args, **kwargs)
    if hasattr(classCall, '__self__'):
      return classCall(*args, **kwargs)
    if callable(classCall):
      return classCall(cls, *args, **kwargs)
    e = typeMsg('classCall', classCall, Callable)
    raise TypeError(e)

  def __instancecheck__(cls, instance) -> bool:
    """This implementation allows the class to customize the instance
    check by implementing a method called __class_instancecheck__."""
    instanceCheck = getattr(cls, '__class_instancecheck__', None)
    if instanceCheck is None:
      return _MetaMetaclass.__instancecheck__(cls, instance)
    if hasattr(instanceCheck, '__self__'):
      return True if instanceCheck(instance) else False
    if callable(instanceCheck):
      return True if instanceCheck(cls, instance) else False
    e = typeMsg('instanceCheck', instanceCheck, Callable)
    raise TypeError(e)

  def __subclasscheck__(cls, subclass) -> bool:
    """This implementation allows the class to customize the subclass
    check by implementing a method called __class_subclasscheck__."""
    subclassCheck = getattr(cls, '__class_subclasscheck__', None)
    if subclassCheck is None:
      return _MetaMetaclass.__subclasscheck__(cls, subclass)
    if hasattr(subclassCheck, '__self__'):
      return True if subclassCheck(subclass) else False
    if callable(subclassCheck):
      return True if subclassCheck(cls, subclass) else False
    e = typeMsg('subclassCheck', subclassCheck, Callable)
    raise TypeError(e)

  def __iter__(cls, ) -> Self:
    """The __iter__ method is invoked to iterate over the class."""
    func = getattr(cls, '__class_iter__', None)
    if func is None:
      func = getattr(_MetaMetaclass, '__iter__')
      return func(cls)
    if hasattr(func, '__self__'):
      return func()
    if callable(func):
      return func(cls)
    e = typeMsg('iter', func, Callable)
    raise TypeError(e)

  def __next__(cls, ) -> Any:
    """The __next__ method is invoked to get the next item in the class."""
    func = getattr(cls, '__class_next__', None)
    if func is None:
      func = getattr(_MetaMetaclass, '__next__')
      return func(cls)
    if hasattr(func, '__self__'):
      return func()
    if callable(func):
      return func(cls)
    e = typeMsg('next', func, Callable)
    raise TypeError(e)



"""AbstractDescriptor provides a common abstract baseclass for all
descriptor classes. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod

from ..waitaminute import MissingVariable

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class _MetaDescriptor(type):
  """MetaDescriptor is a metaclass for the AbstractDescriptor class."""

  def __getitem__(cls, fieldType: type) -> Any:
    """Get the field type."""
    self = cls()
    self._setFieldType(fieldType)
    return self


class AbstractDescriptor(metaclass=_MetaDescriptor):
  """AbstractDescriptor provides a common abstract baseclass for all
  descriptor classes. """

  __field_name__ = None
  __field_owner__ = None

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field and the owner of the field."""
    self.__field_name__ = name
    self.__field_owner__ = owner

  def __get__(self, instance: object, owner: type) -> object:
    """Get the value of the descriptor."""
    if instance is None:
      return self
    return self._instanceGet(instance)

  def __set__(self, instance: object, value: object) -> None:
    """Set the value of the descriptor."""
    self._instanceSet(instance, value)

  def __delete__(self, instance: object) -> None:
    """Delete the value of the descriptor."""
    self._instanceDelete(instance)

  def getFieldName(self, ) -> str:
    """Getter-function for the field name."""
    if self.__field_name__ is None:
      raise MissingVariable('__field_name__', str)
    return self.__field_name__

  def getFieldOwner(self) -> type:
    """Getter-function for the field owner."""
    if self.__field_owner__ is None:
      raise MissingVariable('__field_owner__', type)
    return self.__field_owner__

  @abstractmethod
  def _instanceGet(self, instance: object, **kwargs) -> object:
    """Get the instance of the descriptor."""

  @abstractmethod
  def _instanceSet(self, instance: object, value: object, **kwargs) -> None:
    """Set the instance of the descriptor."""

  @abstractmethod
  def _instanceDelete(self, instance: object, **kwargs) -> None:
    """Delete the instance of the descriptor."""



"""The 'worktoy.attr' module implements the descriptor protocol."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._abstract_descriptor import AbstractDescriptor
from ._abstract_box import AbstractBox
from ._attri_box import AttriBox
from ._flex_box import FlexBox
from ._field import Field
from ._alias import Alias

__all__ = [
    'AbstractDescriptor',
    'AbstractBox',
    'AttriBox',
    'FlexBox',
    'Field',
    'Alias',
]



"""Alias allows a subclass to set an alias for a given attribute inherited
from the parent class. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import Field
from ..mcls import BaseObject
from ..static import overload
from ..text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class Alias(BaseObject):
  """Alias allows a subclass to set an alias for a given attribute inherited
  from the parent class. """

  __field_name__ = None  # new name
  __field_owner__ = None
  __attribute_name__ = None  # original name
  __parent_class__ = None  # original class

  def _getParentDesc(self, ) -> Field:
    """Get the parent descriptor."""
    parent = self.__parent_class__
    origName = self.__attribute_name__
    desc = getattr(parent, origName)
    return desc

  def __set_name__(self, owner: type, name: str) -> None:
    self.__field_owner__ = owner
    self.__field_name__ = name

  def __get__(self, instance: object, owner: type) -> Any:
    """Get the value of the attribute."""
    desc = self._getParentDesc()
    if instance is None:
      return desc
    return desc.__get__(instance, owner)

  def __set__(self, instance: object, value: Any) -> None:
    """Set the value of the attribute."""
    desc = self._getParentDesc()
    return desc.__set__(instance, value)

  def __delete__(self, instance: object) -> None:
    """Delete the value of the attribute."""
    desc = self._getParentDesc()
    return desc.__delete__(instance)

  def __str__(self) -> str:
    """Return the string representation of the Alias object."""
    parentName = self.__parent_class__.__name__
    fieldName = self.__field_name__
    origName = self.__attribute_name__
    ownerName = self.__field_owner__.__name__
    infoSpec = """Alias[new: %s.%s -> old: %s.%s]"""
    return infoSpec % (ownerName, fieldName, parentName, origName)

  def __repr__(self) -> str:
    """Return the string representation of the Alias object."""
    parentName = self.__parent_class__.__name__
    fieldName = self.__field_name__
    origName = self.__attribute_name__
    ownerName = self.__field_owner__.__name__
    infoSpec = """%s.%s = Alias(%s, %s)"""
    return infoSpec % (ownerName, fieldName, parentName, origName)

  @overload(type, str)
  def __init__(self, parent: type, name: str) -> None:
    """Initialize the Alias object."""
    self.__parent_class__ = parent
    self.__attribute_name__ = name

  @overload(str, str)
  def __init__(self, parent: str, name: str) -> None:
    """Initialize the Alias object."""
    cls = type(self)
    for parent in cls.__mro__:
      if parent.__name__ == parent:
        self.__parent_class__ = parent
        self.__attribute_name__ = name
        break
      if parent.__name__ == name:
        self.__parent_class__ = name
        self.__attribute_name__ = parent
        break
    else:
      e = """'%s.__init__' missing required argument: 'parent'!"""
      raise TypeError(monoSpace(e % cls.__name__))



"""BaseObject exposes the functionality of the custom metaclass
implementations in the 'worktoy' library. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import BaseMeta

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False


class BaseObject(metaclass=BaseMeta):
  """BaseObject exposes the functionality of the custom metaclass
  implementations in the 'worktoy' library. """

  def __init__(self, *args, **kwargs) -> None:
    """Why are we still here?"""

  def __init_subclass__(cls, *args, **kwargs) -> None:
    """Just to suffer?"""



"""The Dispatch class dispatches a function call to the appropriate
function based on the type of the first argument. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..static import TypeSig
from ..text import typeMsg, monoSpace
from ..waitaminute import HashMismatch, CastMismatch, ResolveException, \
  VariableNotNone, TypeException
from ..waitaminute import DispatchException

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Callable, TypeAlias, Never

  Types: TypeAlias = tuple[type, ...]
  Hashes: TypeAlias = list[int]
  HashMap: TypeAlias = dict[int, Callable]
  TypesMap: TypeAlias = dict[Types, Callable]
  CastMap: TypeAlias = dict[Types, Callable]
  CallMap: TypeAlias = dict[TypeSig, Callable]


class Dispatch:
  """The Dispatch class dispatches a function call to the appropriate
  function based on the type of the first argument. """

  __field_name__ = None  # name of the function
  __field_owner__ = None  # owner of the function

  __call_map__ = None

  __bound_instance__ = None  # bound instance

  def _getBoundInstance(self) -> object:
    """Get the bound instance."""
    return self.__bound_instance__

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the function."""
    self.__field_name__ = name
    self.__field_owner__ = owner

    for sig, call in self.__call_map__.items():
      if not isinstance(sig, TypeSig):
        raise TypeError(typeMsg('sig', sig, TypeSig))
      if not callable(call):
        from worktoy.mcls import FunctionType
        raise TypeError(typeMsg('call', call, FunctionType))
      TypeSig.replaceTHIS(sig, owner)

  def __init__(self, callMap: CallMap) -> None:
    self.__call_map__ = callMap

  def __get__(self, instance: object, owner: type) -> Dispatch:
    """Get the bound instance."""
    self.__bound_instance__ = instance
    return self

  def __set__(self, *__, **_) -> Never:
    """Set the value of the field."""
    raise TypeError('Cannot set attribute on class')

  def _fastCall(self, *args: Any, **kwargs: Any) -> Any:
    """Fast call the function."""
    instance = self._getBoundInstance()
    for sig, call in self.__call_map__.items():
      try:
        posArgs = sig.fast(*args)
      except HashMismatch:
        continue
      except TypeError as typeError:
        if 'required positional argument' in str(typeError):
          continue
        raise typeError
      if instance is not None:
        return call(instance, *posArgs, **kwargs)
      return call(*posArgs, **kwargs)
    raise DispatchException(self, *args)

  def _flexCall(self, *args: Any, **kwargs: Any) -> Any:
    """Flex call the function."""
    instance = self._getBoundInstance()
    for sig, call in self.__call_map__.items():
      try:
        posArgs = sig.flex(*args)
      except HashMismatch:
        continue
      except CastMismatch:
        continue
      try:
        if instance is not None:
          return call(instance, *posArgs, **kwargs)
        return call(*posArgs, **kwargs)
      except TypeError as typeError:
        continue
    raise DispatchException(self, *args)

  def _resolveArgs(self, *args) -> tuple:
    """Resolves tuples, lists and strings. """
    posArgs = []
    anyResolved = False
    for arg in args:
      if isinstance(arg, list):
        posArgs = [*posArgs, *arg]
        anyResolved = True
        continue
      if isinstance(arg, tuple):
        posArgs = [*posArgs, *arg]
        anyResolved = True
        continue
      if isinstance(arg, str):
        posArgs.append(eval(arg))
        anyResolved = True
        continue
      posArgs.append(arg)
    if anyResolved:
      return (*posArgs,)
    raise ResolveException(self, args)

  def __call__(self, *args: Any, **kwargs: Any) -> Any:
    """Call the function."""
    try:
      return self._fastCall(*args, **kwargs)
    except DispatchException as fastException:
      pass
    try:
      return self._flexCall(*args, **kwargs)
    except DispatchException as flexException:
      pass
    try:
      posArgs = self._resolveArgs(*args)
    except ResolveException as resolveException:
      raise DispatchException(self, *args) from resolveException
    while True:
      try:
        posArgs = self._resolveArgs(*posArgs)
      except ResolveException:
        break
    return self.__call__(*posArgs, **kwargs)

  def getTypeSigs(self) -> list[TypeSig]:
    """Get the type signatures."""
    return [*self.__call_map__.keys(), ]

  def __str__(self, ) -> str:
    """Get the string representation of the function."""
    sigStr = [str(sig) for sig in self.getTypeSigs()]
    info = """%s object supporting type signatures: \n%s"""
    sigLines = '<br><tab>'.join(sigStr)
    return monoSpace(info % (self.__field_name__, sigLines))

  def __repr__(self, ) -> str:
    """Get the string representation of the function."""
    return object.__repr__(self)

  def replaceThis(self, cls: type) -> None:
    """Replace THIS with the class."""
    for sig in self.getTypeSigs():
      sig.replaceTHIS(cls)

  def getFieldName(self, ) -> str:
    """Get the field name."""
    if self.__field_name__ is None:
      raise VariableNotNone('__field_name__')
    if isinstance(self.__field_name__, str):
      return self.__field_name__
    raise TypeException('__field_name__', self.__field_name__, str)

  def getFieldOwner(self, ) -> type:
    """Get the field owner."""
    if self.__field_owner__ is None:
      raise VariableNotNone('__field_owner__')
    if isinstance(self.__field_owner__, type):
      return self.__field_owner__
    raise TypeException('__field_owner__', self.__field_owner__, type)



"""AbstractHook provides an abstract baseclass for hooks used by the
namespaces in the metaclass system. These hooks allow modification of the
class creation process. The following static methods are expected from the
hooks:
- setItem: called before calls to __setitem__ on the namespace object.
- getItem: called before calls to __getitem__ on the namespace object.
- preCompile: called before the final namespace object is populated with
the conventional key, value pairs.
- postCompile: called after the final namespace object is populated with
the conventional key, value pairs.

Subclasses may implement either of the above methods. The default
implementations have no effect, so subclasses need only implement the
methods they are interested in.

AbstractHook implements the descriptor protocol such that calls to
'__get__' receive:
- instance: The current instance of the namespace object.
- owner: The current namespace class

Subclasses should be made available as attributes on the namespace subclass.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...waitaminute import MissingVariable, ReadOnlyError

try:
  from typing import TYPE_CHECKING, Type
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self, Callable, Self, Never
  from worktoy.mcls import AbstractNamespace as ASpace

  AccessorHook = Callable[[ASpace, str, Any], Any]
  CompileHook = Callable[[ASpace, dict], dict]


class AbstractHook:
  """AbstractHook provides an abstract baseclass for hooks used by the
  namespaces in the metaclass system. These hooks allow modification of the
  class creation process. The following static methods are expected from the
  hooks:
  - setItem: called before calls to __setitem__ on the namespace object.
  - getItem: called before calls to __getitem__ on the namespace object.
  - preCompile: called before the final namespace object is populated with
    the conventional key, value pairs.
  - postCompile: called after the final namespace object is populated with
    the conventional key, value pairs.

  Subclasses may implement either of the above methods. The default
  implementations have no effect, so subclasses need only implement the
  methods they are interested in.

  AbstractHook implements the descriptor protocol such that calls to
  '__get__' receive:
  - instance: The current instance of the namespace object.
  - owner: The current namespace class

  Subclasses should be made available as attributes on the namespace
  subclass. """

  __field_name__ = None
  __field_owner__ = None

  __owning_instance__ = None
  __owning_class__ = None

  __owner_hooks_list_name__ = '__hook_objects__'
  __owner_hooks_getter_name__ = 'getHooks'

  @classmethod
  def _getOwnerHooksName(cls) -> str:
    """Getter-function for the name at which the hook is available on the
    namespace class. """
    if cls.__owner_hooks_list_name__ is None:
      raise MissingVariable('__owner_hooks_name__', str)
    return cls.__owner_hooks_list_name__

  @classmethod
  def _getOwnerHooksGetterName(cls) -> str:
    """Getter-function for the name at which the hook is available on the
    namespace class. """
    if cls.__owner_hooks_getter_name__ is None:
      raise MissingVariable('__owner_hooks_getter_name__', str)
    return cls.__owner_hooks_getter_name__

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field and the owner of the field."""
    self.__field_name__ = name
    self.__field_owner__ = owner
    if TYPE_CHECKING:
      assert issubclass(owner, ASpace)
    owner.addHook(self, )

  def __get__(self, instance: object, owner: type) -> Self:
    """Get the wrapped function. The owner is the class through which the
    descriptor is accessed. This is either __field_class__ or a subclass
    hereof. """
    self.__owning_instance__ = instance
    self.__owning_class__ = owner
    return self

  def __set__(self, instance: object, value: object) -> Never:
    """Set the wrapped function. This is not allowed. """
    raise ReadOnlyError(instance, self, value)

  def getOwningNamespace(self) -> ASpace:
    """Get the owning instance. This is the instance of the namespace
    object that owns this hook. """
    if TYPE_CHECKING:
      assert isinstance(self.__owning_instance__, ASpace)
    return self.__owning_instance__

  def getPrimeSpace(self) -> ASpace:
    """Get the prime space. This is the instance of the namespace object
    that owns this hook. """
    namespace = self.getOwningNamespace()
    if TYPE_CHECKING:
      assert isinstance(namespace, ASpace)
    return namespace.getPrimeSpace()

  def getOwningClass(self) -> type:
    """Get the owning class. This is the class of the namespace object
    that owns this hook. """
    return self.__owning_class__

  def getItemHook(self, space: ASpace, key: str, value: object) -> bool:
    """Hook for getItem. This is called before the __getitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False. """
    if TYPE_CHECKING:
      assert isinstance(self, AbstractHook)
    self.__owning_instance__ = space
    return False

  def setItemHook(
      self,
      space: ASpace,
      key: str,
      value: object,
      oldValue: object
  ) -> bool:
    """Hook for setItem. This is called before the __setitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False. """
    if TYPE_CHECKING:
      assert isinstance(self, AbstractHook)
    self.__owning_instance__ = space
    return False

  def preCompileHook(self, space: ASpace, namespace: dict) -> dict:
    """Hook for preCompile. This is called before the __init__ method of
    the namespace object is called. The default implementation does nothing
    and returns the contents unchanged. """
    if TYPE_CHECKING:
      assert isinstance(self, AbstractHook)
    self.__owning_instance__ = space
    return namespace

  def postCompileHook(self, space: ASpace, namespace: dict) -> dict:
    """Hook for postCompile. This is called after the __init__ method of
    the namespace object is called. The default implementation does nothing
    and returns the contents unchanged. """
    if TYPE_CHECKING:
      assert isinstance(self, AbstractHook)
    self.__owning_instance__ = space
    return namespace
